@@ -7,10 +7,14 @@
 #ifndef PDF_ENCRYPT_H
 #define PDF_ENCRYPT_H
 
-#include "PdfDeclarations.h"
 #include "PdfString.h"
 #include "PdfReference.h"
 
+// Define an opaque type for the internal PoDoFo encryption context
+#ifndef PODOFO_CRYPT_CTX
+#define PODOFO_CRYPT_CTX void
+#endif // PODOFO_CRYPT_CTX
+
 namespace PoDoFo
 {
 
@@ -18,8 +22,6 @@ class PdfDictionary;
 class InputStream;
 class PdfObject;
 class OutputStream;
-class AESCryptoEngine;
-class RC4CryptoEngine;
 
 /* Class representing PDF encryption methods. (For internal use only)
  * Based on code from Ulrich Telle: http://wxcode.sourceforge.net/components/wxpdfdoc/
@@ -33,35 +35,32 @@ class RC4CryptoEngine;
  * Licence:     wxWindows licence
  */
 
-/** A enum specifying a valid keylength for a PDF encryption key.
- *  Keys must be in the range 40 to 128 bit and have to be a
- *  multiple of 8.
- *
- *  Adobe Reader supports only keys with 40, 128 or 256 bits!
+/** A enum specifying a valid keylength for a PDF encryption key
  */
-enum class PdfKeyLength
+enum class PdfKeyLength : uint16_t
 {
+    Unknown = 0,
     L40 = 40,
+    L48 = 48,
     L56 = 56,
+    L64 = 64,
+    L72 = 72,
     L80 = 80,
+    L88 = 88,
     L96 = 96,
+    L104 = 104,
+    L112 = 112,
+    L120 = 120,
     L128 = 128,
-#ifdef PODOFO_HAVE_LIBIDN
     L256 = 256
-#endif //PODOFO_HAVE_LIBIDN
 };
 
-#ifdef PODOFO_HAVE_LIBIDN
-enum class PdfAESV3Revision
-{
-    R5 = 5,
-    R6 = 6,
-};
-#endif //PODOFO_HAVE_LIBIDN
-
 /** Set user permissions/restrictions on a document
  */
-enum class PdfPermissions
+// ISO 32000-2:2020 7.6.4.2 "Standard encryption dictionary":
+// "The value of the P entry shall be interpreted as an unsigned
+// 32-bit quantity containing a set of flags."
+enum class PdfPermissions : uint32_t
 {
     None = 0,
     Print = 0x00000004,  ///< Allow printing the document
@@ -85,18 +84,26 @@ enum class PdfPermissions
 /**
  * The encryption algorithm.
  */
-enum class PdfEncryptAlgorithm
+enum class PdfEncryptionAlgorithm : uint8_t
 {
     None = 0,
     RC4V1 = 1,      ///< RC4 Version 1 encryption using a 40bit key
     RC4V2 = 2,      ///< RC4 Version 2 encryption using a key with 40-128bit
     AESV2 = 4,      ///< AES encryption with a 128 bit key (PDF1.6)
-#ifdef PODOFO_HAVE_LIBIDN
-    AESV3 = 8,      ///< AES encryption with a 256 bit key (PDF1.7 extension 3) - Support added by P. Zent
+    AESV3R5 = 8,    ///< AES encryption with a 256 bit key (PDF1.7 extension 3, deprecated in PDF 2.0)
     AESV3R6 = 16,   ///< AES encryption with a 256 bit key, Revision 6 (PDF1.7 extension 8, PDF 2.0)
-#endif //PODOFO_HAVE_LIBIDN
 };
 
+enum class PdfAuthResult : uint8_t
+{
+    Unkwnon = 0,
+    Failed,     ///< Failed to authenticate to this PDF
+    User,       ///< Success authenticating a user for this PDF
+    Owner,      ///< Success authenticating the owner for this PDF
+};
+
+class PdfEncryptContext;
+
 /** A class that is used to encrypt a PDF file and
  *  set document permissions on the PDF file.
  *
@@ -109,6 +116,10 @@ enum class PdfEncryptAlgorithm
  */
 class PODOFO_API PdfEncrypt
 {
+    friend class PdfEncryptMD5Base;
+    friend class PdfEncryptAESV3;
+    PODOFO_PRIVATE_FRIEND(class PdfEncryptSession);
+
 public:
     virtual ~PdfEncrypt();
 
@@ -123,14 +134,14 @@ public:
      *  \param keyLength the length of the encryption key ranging from 40 to 128 bits
      *                    (only used if algorithm == PdfEncryptAlgorithm::RC4V2)
      *
-     *  \see GenerateEncryptionKey with the documentID to generate the real
+     *  \see EnsureEncryptionInitialized with the documentID to generate the real
      *       encryption key using this information
      */
     static std::unique_ptr<PdfEncrypt> Create(const std::string_view& userPassword,
         const std::string_view& ownerPassword,
         PdfPermissions protection = PdfPermissions::Default,
-        PdfEncryptAlgorithm algorithm = PdfEncryptAlgorithm::AESV2,
-        PdfKeyLength keyLength = PdfKeyLength::L40);
+        PdfEncryptionAlgorithm algorithm = PdfEncryptionAlgorithm::AESV3R6,
+        PdfKeyLength keyLength = PdfKeyLength::Unknown);
 
     /** Initialize a PdfEncrypt object from an encryption dictionary in a PDF file.
      *
@@ -145,12 +156,6 @@ public:
      */
     static std::unique_ptr<PdfEncrypt> CreateFromObject(const PdfObject& obj);
 
-    /** Copy constructor
-     *
-     *  \param rhs another PdfEncrypt object which is copied
-     */
-    static std::unique_ptr<PdfEncrypt> CreateFromEncrypt(const PdfEncrypt& rhs);
-
     /**
      * Retrieve the list of encryption algorithms that are used
      * when loading a PDF document.
@@ -162,19 +167,7 @@ public:
      *
      * \return an or'ed together list of all enabled encryption algorithms
      */
-    static PdfEncryptAlgorithm GetEnabledEncryptionAlgorithms();
-
-    /**
-     * Specify the list of encryption algorithms that should be used by PoDoFo
-     * when loading a PDF document.
-     *
-     * This can be used to disable for example AES encryption/decryption
-     * which is unstable in certain cases.
-     *
-     * \see GetEnabledEncryptionAlgorithms
-     * \see IsEncryptionEnabled
-     */
-    static void SetEnabledEncryptionAlgorithms(PdfEncryptAlgorithm nEncryptionAlgorithms);
+    static PdfEncryptionAlgorithm GetEnabledEncryptionAlgorithms();
 
     /**
      * Test if a certain encryption algorithm is enabled for loading PDF documents.
@@ -183,13 +176,23 @@ public:
      * \see GetEnabledEncryptionAlgorithms
      * \see SetEnabledEncryptionAlgorithms
      */
-    static bool IsEncryptionEnabled(PdfEncryptAlgorithm algorithm);
+    static bool IsEncryptionEnabled(PdfEncryptionAlgorithm algorithm);
 
-    /** Generate encryption key from user and owner passwords and protection key
+    /** Ensure encryption key and /O, /U, /OE, /UE values are initialized
      *
      *  \param documentId the documentId of the current document
      */
-    void GenerateEncryptionKey(const PdfString& documentId);
+    void EnsureEncryptionInitialized(const PdfString& documentId, PdfEncryptContext& context);
+
+    /**
+     * Tries to authenticate a user using either the user or owner password
+     *
+     * \param password owner or user password
+     * \param documentId the documentId of the PDF file
+     *
+     * \returns true if either the owner or user password matches password
+     */
+    void Authenticate(const std::string_view& password, const PdfString& documentId, PdfEncryptContext& context) const;
 
     /** Fill all keys into a encryption dictionary.
      *  This dictionary is usually added to the PDF files trailer
@@ -210,7 +213,8 @@ public:
      *
      *  \returns an InputStream that decrypts all data.
      */
-    virtual std::unique_ptr<InputStream> CreateEncryptionInputStream(InputStream& inputStream, size_t inputLen, const PdfReference& objref) = 0;
+    virtual std::unique_ptr<InputStream> CreateEncryptionInputStream(InputStream& inputStream, size_t inputLen,
+        PdfEncryptContext& context, const PdfReference& objref) const = 0;
 
     /** Create an OutputStream that encrypts all data written to
      *  it using the current settings of the PdfEncrypt object.
@@ -222,22 +226,13 @@ public:
      *
      *  \returns a OutputStream that encrypts all data.
      */
-    virtual std::unique_ptr<OutputStream> CreateEncryptionOutputStream(OutputStream& outputStream, const PdfReference& objref) = 0;
-
-    /**
-     * Tries to authenticate a user using either the user or owner password
-     *
-     * \param password owner or user password
-     * \param documentId the documentId of the PDF file
-     *
-     * \returns true if either the owner or user password matches password
-     */
-    bool Authenticate(const std::string_view& password, const PdfString& documentId);
+    virtual std::unique_ptr<OutputStream> CreateEncryptionOutputStream(OutputStream& outputStream,
+        PdfEncryptContext& context, const PdfReference& objref) const = 0;
 
     /** Get the encryption algorithm of this object.
      * \returns the PdfEncryptAlgorithm of this object
      */
-    inline PdfEncryptAlgorithm GetEncryptAlgorithm() const { return m_Algorithm; }
+    inline PdfEncryptionAlgorithm GetEncryptAlgorithm() const { return m_Algorithm; }
 
     /** Checks if an owner password is set.
      *  An application reading PDF must adhere to permissions for printing,
@@ -248,8 +243,7 @@ public:
      *
      *  \see PdfEncrypt to set own document permissions.
      */
-    inline bool IsOwnerPasswordSet() const { return !m_ownerPass.empty(); }
-
+    bool IsOwnerPasswordSet() const;
 
     /** Checks if printing this document is allowed.
      *  Every PDF consuming applications has to adhere this value!
@@ -323,41 +317,13 @@ public:
      */
     bool IsHighPrintAllowed() const;
 
-    /** Get the U object value (user)
-     */
-    const unsigned char* GetUValue() const { return m_uValue; }
-
-    /** Get the O object value (owner)
-     */
-    const unsigned char* GetOValue() const { return m_oValue; }
-
-    /** Get the encryption key value (owner)
-     */
-    const unsigned char* GetEncryptionKey() const { return m_encryptionKey; }
-
-    /** Get the P object value (protection)
-     */
-    PdfPermissions GetPValue() const { return m_pValue; }
-
-    /** Get the revision number of the encryption method
-     */
-    int GetRevision() const { return m_rValue; }
-
-    /** Get the key length of the encryption key in bits
-     */
-    int GetKeyLength() const;
-
-    /** Is metadata encrypted
-     */
-    bool IsMetadataEncrypted() const { return m_EncryptMetadata; }
-
     /** Encrypt a character span
      */
-    void EncryptTo(charbuff& out, const bufferview& view, const PdfReference& objref) const;
+    void EncryptTo(charbuff& out, const bufferview& view, PdfEncryptContext& context, const PdfReference& objref) const;
 
     /** Decrypt a character span
      */
-    void DecryptTo(charbuff& out, const bufferview& view, const PdfReference& objref) const;
+    void DecryptTo(charbuff& out, const bufferview& view, PdfEncryptContext& context, const PdfReference& objref) const;
 
     /** Calculate stream size
      */
@@ -367,104 +333,170 @@ public:
      */
     virtual size_t CalculateStreamOffset() const = 0;
 
-protected:
-    PdfEncrypt();
+    /** Get the encryption key length in bytes
+     * \remarks The maximum is 32 bytes
+     */
+    unsigned GetKeyLengthBytes() const;
 
-    PdfEncrypt(const PdfEncrypt& rhs);
+    /** Get the U object value (user)
+     */
+    bufferview GetUValue() const;
+
+    /** Get the O object value (owner)
+     */
+    bufferview GetOValue() const;
+
+public:
+    /** Get the length of the encryption key in bits
+     */
+    PdfKeyLength GetKeyLength() const { return m_KeyLength; }
+
+    /** Get the P object value (protection)
+     */
+    inline PdfPermissions GetPValue() const { return m_pValue; }
+
+    /** Get the revision number of the encryption method
+     */
+    inline unsigned GetRevision() const { return m_rValue; }
+
+    inline bool IsMetadataEncrypted() const { return m_EncryptMetadata; }
+
+    inline bool IsParsed() const { return m_IsParsed; }
 
 protected:
-    virtual void Decrypt(const char* inStr, size_t inLen, const PdfReference& objref,
-        char* outStr, size_t& outLen) const = 0;
+    inline const unsigned char* GetUValueRaw() const { return m_uValue; }
 
-    virtual void Encrypt(const char* inStr, size_t inLen, const PdfReference& objref,
-        char* outStr, size_t outLen) const = 0;
+    inline const unsigned char* GetOValueRaw() const { return m_oValue; }
 
-    virtual bool Authenticate(const std::string_view& password, const  std::string_view& documentId) = 0;
+    inline const std::string& GetUserPassword() const { return m_userPass; }
 
-    virtual void GenerateEncryptionKey(const std::string_view& documentId) = 0;
+    inline const std::string& GetOwnerPassword() const { return m_ownerPass; }
+
+    int64_t GetPValueForSerialization() const;
+
+protected:
+    /**
+     * Init from parsed values
+     */
+    void InitFromValues(PdfEncryptionAlgorithm algorithm, PdfKeyLength keyLength, unsigned char revision,
+        PdfPermissions pValue, const bufferview& uValue, const bufferview& oValue,
+        bool encryptedMetadata);
+
+    /**
+     * Init from scratch with user/owner password. Requires full
+     * initialization with EnsureEncryptionInitialized
+     */
+    void InitFromScratch(const std::string_view& userPassword, const std::string_view& ownerPassword,
+        PdfEncryptionAlgorithm algorithm, PdfKeyLength keyLength, unsigned char revision,
+        PdfPermissions pValue, bool encryptedMetadata);
+
+    virtual void Decrypt(const char* inStr, size_t inLen, PdfEncryptContext& context,
+        const PdfReference& objref, char* outStr, size_t& outLen) const = 0;
+
+    virtual void Encrypt(const char* inStr, size_t inLen, PdfEncryptContext& context,
+        const PdfReference& objref, char* outStr, size_t outLen) const = 0;
+
+    virtual PdfAuthResult Authenticate(const std::string_view& password, const std::string_view& documentId,
+        PODOFO_CRYPT_CTX* ctx, unsigned char encryptionKey[32]) const = 0;
+
+    virtual void GenerateEncryptionKey(
+        const std::string_view& documentId, PdfAuthResult authResult, PODOFO_CRYPT_CTX* ctx,
+        unsigned char uValue[48], unsigned char oValue[48], unsigned char encryptionKey[32]) = 0;
 
     // Check two keys for equality
-    bool CheckKey(unsigned char key1[32], unsigned char key2[32]);
+    bool CheckKey(const unsigned char key1[32], const unsigned char key2[32]) const;
 
-    PdfEncryptAlgorithm m_Algorithm;   // The used encryption algorithm
-    PdfKeyLength m_eKeyLength;         // The encryption key length, as enum value
-    unsigned m_keyLength;              // Length of encryption key
-    int m_rValue;                      // Revision
-    PdfPermissions m_pValue;           // P entry in pdf document
-    std::string m_userPass;            // User password
-    std::string m_ownerPass;           // Owner password
-    unsigned char m_uValue[48];        // U entry in pdf document
-    unsigned char m_oValue[48];        // O entry in pdf document
-    unsigned char m_encryptionKey[32]; // Encryption key
-    std::string m_documentId;          // DocumentID of the current document
-    bool m_EncryptMetadata;            // Is metadata encrypted
+    enum class PdfRC4Revision : uint8_t
+    {
+        R2 = 2,
+        R3 = 3,
+    };
+
+    enum class PdfAESV3Revision : uint8_t
+    {
+        R5 = 5,
+        R6 = 6,
+    };
 
 private:
-    static PdfEncryptAlgorithm s_EnabledEncryptionAlgorithms; // Or'ed int containing the enabled encryption algorithms
+    PdfEncrypt();
+
+    PdfEncrypt(const PdfEncrypt& rhs) = default;
+
+    PdfEncrypt& operator=(PdfEncrypt& rhs) = delete;
+
+private:
+    static std::unique_ptr<PdfEncrypt> CreateFromEncrypt(const PdfEncrypt& rhs);
+
+    void clearSensitiveInfo();
+
+private:
+    PdfEncryptionAlgorithm m_Algorithm;   // The used encryption algorithm
+    unsigned char m_rValue;            // Revision
+    PdfKeyLength m_KeyLength;          // The encryption key length, as enum value
+    PdfPermissions m_pValue;           // P entry in pdf document
+    unsigned char m_uValue[48];        // U entry in pdf document
+    unsigned char m_oValue[48];        // O entry in pdf document
+    unsigned char m_uValueSize;
+    unsigned char m_oValueSize;
+    bool m_EncryptMetadata;            // Is metadata encrypted
+    bool m_IsParsed;                   // True if the object is created from parsed values
+    bool m_initialized;               // True if the object O/U values were filled
+    std::string m_userPass;            // User password
+    std::string m_ownerPass;           // Owner password
+
 };
 
-#ifdef PODOFO_HAVE_LIBIDN
-
-/** A pure virtual class that is used to encrypt a PDF file (AES-256)
- *  This class is the base for classes that implement algorithms based on SHA hashes
- *
- *  Client code is working only with PdfEncrypt class and knows nothing
- *	about PdfEncrypt*, it is created through CreatePdfEncrypt factory method
- *
- */
-class PdfEncryptSHABase : public PdfEncrypt
+class PODOFO_API PdfEncryptContext final
 {
+    friend class PdfEncrypt;
+    friend class PdfEncryptRC4;
+    friend class PdfEncryptAESV2;
+    friend class PdfEncryptAESV3;
+
 public:
-    PdfEncryptSHABase();
+    PdfEncryptContext();
 
-    // copy constructor
-    PdfEncryptSHABase(const PdfEncrypt& rhs);
+    ~PdfEncryptContext();
 
-    void CreateEncryptionDictionary(PdfDictionary& dictionary) const override;
+    PdfEncryptContext(const PdfEncryptContext&);
 
-    // Get the UE object value (user)
-    const unsigned char* GetUEValue() const { return m_ueValue; }
+    PdfEncryptContext& operator=(const PdfEncryptContext&);
 
-    // Get the OE object value (owner)
-    const unsigned char* GetOEValue() const { return m_oeValue; }
+public:
+    inline PdfAuthResult GetAuthResult() { return m_AuthResult; }
 
-    // Get the Perms object value (encrypted protection)
-    const unsigned char* GetPermsValue() const { return m_permsValue; }
+    inline const std::string GetDocumentId() { return m_documentId; }
 
-    bool Authenticate(const std::string_view& documentID, const std::string_view& password,
-        const bufferview& uValue, const std::string_view& ueValue,
-        const bufferview& oValue, const std::string_view& oeValue,
-        PdfPermissions pValue, const std::string_view& permsValue,
-        int lengthValue, int rValue);
+    bool IsAuthenticated() const;
 
-protected:
-    // NOTE: We must declare again without body otherwise the other Authenticate overload hides it
-    bool Authenticate(const std::string_view& password, const std::string_view& documentId) override = 0;
+private:
+    inline const unsigned char* GetEncryptionKey() const { return m_encryptionKey; }
 
-    // Generate initial vector
-    void GenerateInitialVector(unsigned char iv[]) const;
+    PODOFO_CRYPT_CTX* GetCryptCtx();
 
-    // Compute encryption key to be used with AES-256
-    void ComputeEncryptionKey();
+    template <typename T>
+    T& GetCustomCtx()
+    {
+        if (m_customCtx == nullptr)
+        {
+            m_customCtx = ::operator new(sizeof(T));
+            m_customCtxSize = sizeof(T);
+        }
 
-    // Compute hash for password and salt with optional uValue
-    void ComputeHash(const unsigned char* pswd, unsigned pswdLen, unsigned char salt[8], unsigned char uValue[48], unsigned char hashValue[32]);
+        return *(T*)m_customCtx;
+    }
 
-    // Generate the U and UE entries
-    void ComputeUserKey(const unsigned char* userpswd, unsigned len);
-
-    // Generate the O and OE entries
-    void ComputeOwnerKey(const unsigned char* userpswd, unsigned len);
-
-    // Preprocess password for use in EAS-256 Algorithm
-    // outBuf needs to be at least 127 bytes long
-    void PreprocessPassword(const std::string_view& password, unsigned char* outBuf, unsigned& len);
-
-    unsigned char m_ueValue[32];        // UE entry in pdf document
-    unsigned char m_oeValue[32];        // OE entry in pdf document
-    unsigned char m_permsValue[16];     // Perms entry in pdf document
+private:
+    unsigned char m_encryptionKey[32]; // Encryption key
+    std::string m_documentId;          // DocumentID of the current document
+    PdfAuthResult m_AuthResult;
+    PODOFO_CRYPT_CTX* m_cryptCtx;
+    void* m_customCtx;
+    size_t m_customCtxSize;
 };
 
+
 /** A pure virtual class that is used to encrypt a PDF file (RC4, AES-128)
  *  This class is the base for classes that implement algorithms based on MD5 hashes
  *
@@ -472,112 +504,45 @@ protected:
  *	about PdfEncrypt*, it is created through CreatePdfEncrypt factory method
  *
  */
-#endif // PODOFO_HAVE_LIBIDN
-
-/** A pure virtual class that is used to encrypt a PDF file (AES-128/256)
- *  This class is the base for classes that implement algorithms based on AES
- *
- *  Client code is working only with PdfEncrypt class and knows nothing
- *	about PdfEncrypt*, it is created through CreatePdfEncrypt factory method
- *
- */
-class PdfEncryptAESBase
+class PdfEncryptMD5Base : public PdfEncrypt
 {
-public:
-    ~PdfEncryptAESBase();
+    friend class PdfEncryptRC4;
+    friend class PdfEncryptAESV2;
 
-protected:
-    PdfEncryptAESBase();
-
-    void BaseDecrypt(const unsigned char* key, unsigned keylen, const unsigned char* iv,
-        const unsigned char* textin, size_t textlen,
-        unsigned char* textout, size_t& textoutlen) const;
-    void BaseEncrypt(const unsigned char* key, unsigned keylen, const unsigned char* iv,
-        const unsigned char* textin, size_t textlen,
-        unsigned char* textout, size_t textoutlen) const;
-
-    AESCryptoEngine* m_aes;                // AES encryptor
-};
-
-/** A pure virtual class that is used to encrypt a PDF file (RC4-40..128)
- *  This class is the base for classes that implement algorithms based on RC4
- *
- *  Client code is working only with PdfEncrypt class and knows nothing
- *	about PdfEncrypt*, it is created through CreatePdfEncrypt factory method
- *
- */
-class PdfEncryptRC4Base
-{
-public:
-    ~PdfEncryptRC4Base();
-
-protected:
-    PdfEncryptRC4Base();
-
-    // AES encryption
-    void RC4(const unsigned char* key, unsigned keylen,
-        const unsigned char* textin, size_t textlen,
-        unsigned char* textout, size_t textoutlen) const;
-
-    RC4CryptoEngine* m_rc4;                // AES encryptor
-};
-
-class PdfEncryptMD5Base : public PdfEncrypt, public PdfEncryptRC4Base
-{
-public:
+private:
     PdfEncryptMD5Base();
 
-    // copy constructor
-    PdfEncryptMD5Base(const PdfEncrypt& rhs);
+    PdfEncryptMD5Base(const PdfEncryptMD5Base& rhs);
 
+public:
     void CreateEncryptionDictionary(PdfDictionary& dictionary) const override;
 
-    /** Create a PdfString of MD5 data generated from a buffer in memory.
-     *  \param buffer the buffer of which to calculate the MD5 sum
-     *  \param length the length of the buffer
-     *
-     *  \returns an MD5 sum as PdfString
-     */
-    static PdfString GetMD5String(const unsigned char* buffer, unsigned length);
-
-    // Calculate the binary MD5 message digest of the given data
-    static void GetMD5Binary(const unsigned char* data, unsigned length, unsigned char* digest);
-
     // NOTE: We must declare again without body otherwise the other Authenticate overload hides it
-    bool Authenticate(const std::string_view& password, const std::string_view& documentId) override = 0;
-
-    bool Authenticate(const std::string_view& documentID, const std::string_view& password,
-        const bufferview& uValue, const bufferview& oValue,
-        PdfPermissions pValue, int lengthValue, int rValue);
+    PdfAuthResult Authenticate(const std::string_view& password, const std::string_view& documentId,
+        PODOFO_CRYPT_CTX* ctx, unsigned char encryptionKey[32]) const override = 0;
 
 protected:
-
-    // Generate initial vector
-    void GenerateInitialVector(unsigned char iv[]) const;
-
     // Compute owner key
-    void ComputeOwnerKey(const unsigned char userPad[32], const unsigned char ownerPad[32],
-        int keylength, int revision, bool authenticate, unsigned char ownerKey[32]);
+    static void ComputeOwnerKey(const unsigned char userPad[32], const unsigned char ownerPad[32],
+        unsigned keylength, unsigned revision, bool authenticate, PODOFO_CRYPT_CTX* ctx, unsigned char ownerKey[32]);
 
     // Pad a password to 32 characters
-    void PadPassword(const std::string_view& password, unsigned char pswd[32]);
+    static void PadPassword(const std::string_view& password, unsigned char pswd[32]);
 
     // Compute encryption key and user key
-    void ComputeEncryptionKey(const std::string_view& documentID,
+    static void ComputeEncryptionKey(const std::string_view& documentID,
         const unsigned char userPad[32], const unsigned char ownerKey[32],
-        PdfPermissions pValue, PdfKeyLength keyLength, int revision,
-        unsigned char userKey[32], bool encryptMetadata);
+        PdfPermissions pValue, unsigned keyLength, unsigned revision,
+        bool encryptMetadata, PODOFO_CRYPT_CTX* ctx,
+        unsigned char userKey[32], unsigned char encryptionKey[32]);
 
     /** Create the encryption key for the current object.
      *
      *  \param objkey pointer to an array of at least MD5_HASHBYTES (=16) bytes length
      *  \param pnKeyLen pointer to an integer where the actual keylength is stored.
      */
-    void CreateObjKey(unsigned char objkey[16], unsigned& pnKeyLen, const PdfReference& objref) const;
-
-    unsigned char m_rc4key[16];         // last RC4 key
-    unsigned char m_rc4last[256];       // last RC4 state table
-
+    void CreateObjKey(unsigned char objkey[16], unsigned& pnKeyLen,
+        const unsigned char m_encryptionKey[32], const PdfReference& objref) const;
 };
 
 /** A class that is used to encrypt a PDF file (AES-128)
@@ -586,109 +551,174 @@ protected:
  *	about PdfEncryptAES*, it is created through CreatePdfEncrypt factory method
  *
  */
-class PdfEncryptAESV2 : public PdfEncryptMD5Base, public PdfEncryptAESBase
+class PdfEncryptAESV2 final : public PdfEncryptMD5Base
 {
-public:
-    PdfEncryptAESV2(PdfString oValue, PdfString uValue, PdfPermissions pValue, bool bEncryptMetadata);
+    friend class PdfEncrypt;
+
+private:
+    PdfEncryptAESV2(PdfString oValue, PdfString uValue, PdfPermissions pValue, bool encryptMetadata);
     PdfEncryptAESV2(const std::string_view& userPassword, const std::string_view& ownerPassword,
-        PdfPermissions protection = PdfPermissions::Default);
-    PdfEncryptAESV2(const PdfEncrypt& rhs);
+        PdfPermissions protection);
+    PdfEncryptAESV2(const PdfEncryptAESV2& rhs);
 
-    std::unique_ptr<InputStream> CreateEncryptionInputStream(InputStream& inputStream, size_t inputLen, const PdfReference& objref) override;
-    std::unique_ptr<OutputStream> CreateEncryptionOutputStream(OutputStream& outputStream, const PdfReference& objref) override;
-
-    void Encrypt(const char* inStr, size_t inLen, const PdfReference& objref,
-        char* outStr, size_t outLen) const override;
-    void Decrypt(const char* inStr, size_t inLen, const PdfReference& objref,
-        char* outStr, size_t& outLen) const override;
+public:
+    std::unique_ptr<InputStream> CreateEncryptionInputStream(InputStream& inputStream, size_t inputLen,
+        PdfEncryptContext& context, const PdfReference& objref) const override;
+    std::unique_ptr<OutputStream> CreateEncryptionOutputStream(OutputStream& outputStream,
+        PdfEncryptContext& context, const PdfReference& objref) const override;
 
     size_t CalculateStreamOffset() const override;
 
     size_t CalculateStreamLength(size_t length) const override;
 
 protected:
-    void GenerateEncryptionKey(const std::string_view& documentId) override;
+    void Encrypt(const char* inStr, size_t inLen, PdfEncryptContext& context,
+        const PdfReference& objref, char* outStr, size_t outLen) const override;
+    void Decrypt(const char* inStr, size_t inLen, PdfEncryptContext& context,
+        const PdfReference& objref, char* outStr, size_t& outLen) const override;
 
-    bool Authenticate(const std::string_view& password, const std::string_view& documentId) override;
+    void GenerateEncryptionKey(const std::string_view& documentId, PdfAuthResult authResult, PODOFO_CRYPT_CTX* ctx,
+        unsigned char uValue[48], unsigned char oValue[48], unsigned char encryptionKey[32]) override;
+
+    PdfAuthResult Authenticate(const std::string_view& password, const std::string_view& documentId,
+        PODOFO_CRYPT_CTX* ctx, unsigned char encryptionKey[32]) const override;
+
+private:
+    void generateInitialVector(const std::string_view& documentId, unsigned char iv[]) const;
 };
 
-#ifdef PODOFO_HAVE_LIBIDN
-
 /** A class that is used to encrypt a PDF file (AES-256)
  *
  *  Client code is working only with PdfEncrypt class and knows nothing
  *	about PdfEncryptAES*, it is created through CreatePdfEncrypt factory method
  *
  */
-class PdfEncryptAESV3 : public PdfEncryptSHABase, public PdfEncryptAESBase
+class PdfEncryptAESV3 final : public PdfEncrypt
 {
-public:
+    friend class PdfEncrypt;
+
+private:
     PdfEncryptAESV3(PdfString oValue, PdfString oeValue, PdfString uValue, PdfString ueValue,
         PdfPermissions pValue, PdfString permsValue, PdfAESV3Revision rev);
     PdfEncryptAESV3(const std::string_view& userPassword, const std::string_view& ownerPassword,
-        PdfAESV3Revision rev, PdfPermissions protection = PdfPermissions::Default);
-    PdfEncryptAESV3(const PdfEncrypt& rhs);
+        PdfAESV3Revision rev, PdfPermissions protection);
+    PdfEncryptAESV3(const PdfEncryptAESV3& rhs);
 
-    std::unique_ptr<InputStream> CreateEncryptionInputStream(InputStream& inputStream, size_t inputLen, const PdfReference& objref) override;
-    std::unique_ptr<OutputStream> CreateEncryptionOutputStream(OutputStream& outputStream, const PdfReference& objref) override;
-
-    // Encrypt a character string
-    void Encrypt(const char* inStr, size_t inLen, const PdfReference& objref,
-        char* outStr, size_t outLen) const override;
-    void Decrypt(const char* inStr, size_t inLen, const PdfReference& objref,
-        char* outStr, size_t& outLen) const override;
+public:
+    std::unique_ptr<InputStream> CreateEncryptionInputStream(InputStream& inputStream, size_t inputLen,
+        PdfEncryptContext& context, const PdfReference& objref) const override;
+    std::unique_ptr<OutputStream> CreateEncryptionOutputStream(OutputStream& outputStream,
+        PdfEncryptContext& context, const PdfReference& objref) const override;
 
     size_t CalculateStreamOffset() const override;
 
     size_t CalculateStreamLength(size_t length) const override;
 
+    void CreateEncryptionDictionary(PdfDictionary& dictionary) const override;
+
+    // Get the UE object value (user)
+    bufferview GetUEValue() const;
+
+    // Get the OE object value (owner)
+    bufferview GetOEValue() const;
+
+    // Get the Perms object value (encrypted protection)
+    bufferview GetPermsValue() const;
+
 protected:
-    bool Authenticate(const std::string_view& password, const std::string_view& documentId) override;
+    // Encrypt a character string
+    void Encrypt(const char* inStr, size_t inLen, PdfEncryptContext& context,
+        const PdfReference& objref, char* outStr, size_t outLen) const override;
+    void Decrypt(const char* inStr, size_t inLen, PdfEncryptContext& context,
+        const PdfReference& objref, char* outStr, size_t& outLen) const override;
 
-    void GenerateEncryptionKey(const std::string_view& documentId) override;
+    PdfAuthResult Authenticate(const std::string_view& password, const std::string_view& documentId,
+        PODOFO_CRYPT_CTX* ctx, unsigned char encryptionKey[32]) const override;
+
+    void GenerateEncryptionKey(const std::string_view& documentId, PdfAuthResult authResult, PODOFO_CRYPT_CTX* ctx,
+        unsigned char uValue[48], unsigned char oValue[48], unsigned char encryptionKey[32]) override;
+
+private:
+    // Generate initial vector
+    static void generateInitialVector(unsigned char iv[]);
+
+    // Preprocess password for use in EAS-256 Algorithm
+    // outBuf needs to be at least 127 bytes long
+    static void preprocessPassword(const std::string_view& password, unsigned char* outBuf, unsigned& len);
+
+    // Compute encryption key to be used with AES-256
+    static void computeEncryptionKey(unsigned keyLength, unsigned char encryptionKey[32]);
+
+    // Compute hash for password and salt with optional uValue
+    static void computeHash(const unsigned char* pswd, unsigned pswdLen, unsigned revision,
+        const unsigned char salt[8], const unsigned char uValue[48], unsigned char hashValue[32]);
+
+    // Generate the U and UE entries
+    static void computeUserKey(const unsigned char* userpswd, unsigned len, unsigned revision,
+        unsigned keyLength, const unsigned char encryptionKey[32],
+        unsigned char uValue[48], unsigned char ueValue[32]);
+
+    // Generate the O and OE entries
+    static void computeOwnerKey(const unsigned char* userpswd, unsigned len, unsigned revision,
+        unsigned keyLength, const unsigned char encryptionKey[32], const unsigned char uValue[48],
+        unsigned char oValue[48], unsigned char oeValue[32]);
+
+private:
+    unsigned char m_ueValue[32];        // UE entry in pdf document
+    unsigned char m_oeValue[32];        // OE entry in pdf document
+    unsigned char m_permsValue[16];     // Perms entry in pdf document
 };
 
-#endif // PODOFO_HAVE_LIBIDN
-
 /** A class that is used to encrypt a PDF file (RC4 40-bit and 128-bit)
  *
  *  Client code is working only with PdfEncrypt class and knows nothing
  *	about PdfEncryptRC4, it is created through CreatePdfEncrypt factory method
  *
  */
-class PdfEncryptRC4 : public PdfEncryptMD5Base
+class PdfEncryptRC4 final : public PdfEncryptMD5Base
 {
-public:
-    PdfEncryptRC4(PdfString oValue, PdfString uValue,
-        PdfPermissions pValue, int rValue, PdfEncryptAlgorithm algorithm, int length, bool encryptMetadata);
+    friend class PdfEncrypt;
+
+private:
+    PdfEncryptRC4(PdfString oValue, PdfString uValue, PdfPermissions pValue,
+        PdfRC4Revision revision, PdfEncryptionAlgorithm algorithm,
+        unsigned keyLength, bool encryptMetadata);
     PdfEncryptRC4(const std::string_view& userPassword, const std::string_view& ownerPassword,
-        PdfPermissions protection = PdfPermissions::Default,
-        PdfEncryptAlgorithm algorithm = PdfEncryptAlgorithm::RC4V1,
-        PdfKeyLength keyLength = PdfKeyLength::L40);
-    PdfEncryptRC4(const PdfEncrypt& rhs);
+        PdfPermissions protection,
+        PdfEncryptionAlgorithm algorithm,
+        PdfKeyLength keyLength);
+    PdfEncryptRC4(const PdfEncryptRC4& rhs) = default;
 
-    void Encrypt(const char* inStr, size_t inLen, const PdfReference& objref,
-        char* outStr, size_t outLen) const override;
+public:
+    std::unique_ptr<InputStream> CreateEncryptionInputStream(InputStream& inputStream, size_t inputLen,
+        PdfEncryptContext& context, const PdfReference& objref) const override;
 
-    void Decrypt(const char* inStr, size_t inLen, const PdfReference& objref,
-        char* outStr, size_t& outLen) const override;
-
-    std::unique_ptr<InputStream> CreateEncryptionInputStream(InputStream& inputStream, size_t inputLen, const PdfReference& objref) override;
-
-    std::unique_ptr<OutputStream> CreateEncryptionOutputStream(OutputStream& outputStream, const PdfReference& objref) override;
+    std::unique_ptr<OutputStream> CreateEncryptionOutputStream(OutputStream& outputStream,
+        PdfEncryptContext& context, const PdfReference& objref) const override;
 
     size_t CalculateStreamOffset() const override;
 
     size_t CalculateStreamLength(size_t length) const override;
 
 protected:
-    void GenerateEncryptionKey(const std::string_view& documentId) override;
+    void Encrypt(const char* inStr, size_t inLen, PdfEncryptContext& context,
+        const PdfReference& objref, char* outStr, size_t outLen) const override;
 
-    bool Authenticate(const std::string_view& password, const std::string_view& documentId) override;
+    void Decrypt(const char* inStr, size_t inLen, PdfEncryptContext& context,
+        const PdfReference& objref, char* outStr, size_t& outLen) const override;
+
+    void GenerateEncryptionKey(const std::string_view& documentId, PdfAuthResult authResult, PODOFO_CRYPT_CTX* ctx,
+        unsigned char uValue[48], unsigned char oValue[48], unsigned char encryptionKey[32]) override;
+
+    PdfAuthResult Authenticate(const std::string_view& password, const std::string_view& documentId,
+        PODOFO_CRYPT_CTX* ctx, unsigned char encryptionKey[32]) const override;
+
+private:
+    static unsigned normalizeKeyLength(unsigned keyLength);
 };
 
 }
 ENABLE_BITMASK_OPERATORS(PoDoFo::PdfPermissions);
-ENABLE_BITMASK_OPERATORS(PoDoFo::PdfEncryptAlgorithm);
+ENABLE_BITMASK_OPERATORS(PoDoFo::PdfEncryptionAlgorithm);
 
 #endif // PDF_ENCRYPT_H
