@@ -16,11 +16,13 @@ namespace PoDoFo {
 /** This is a common base class for simple, non CID-keyed fonts
  * like Type1, TrueType and Type3
  */
-class PdfFontSimple : public PdfFont
+class PODOFO_API PdfFontSimple : public PdfFont
 {
-    friend class PdfFontStandard14;
+    friend class PdfFontTrueType;
+    friend class PdfFontType1;
+    friend class PdfFontType3;
 
-protected:
+private:
     /** Create a new PdfFont object which will introduce itself
      *  automatically to every page object it is used on.
      *
@@ -32,19 +34,18 @@ protected:
      *                   depending on pEncoding->IsAutoDelete()
      *
      */
-    PdfFontSimple(PdfDocument& doc, const PdfFontMetricsConstPtr& metrics,
-        const PdfEncoding& encoding);
+    PdfFontSimple(PdfDocument& doc, PdfFontType type,
+        PdfFontMetricsConstPtr&& metrics, const PdfEncoding& encoding);
 
 protected:
-    void Init();
-
     void embedFont() override final;
 
+    void embedFontSubset() override final;
+
     void initImported() override;
 
 private:
     void getWidthsArray(PdfArray& widths) const;
-    void getFontMatrixArray(PdfArray& fontMatrix) const;
 
 protected:
     PdfObject* m_Descriptor;
